"""Helper functions for the project"""
import re
from configparser import ConfigParser
from decimal import Decimal
from pathlib import Path
from typing import Dict, List, Union

from .exceptions import DecodeError

SET_EXPORT_EXPRESSION = re.compile(r'[\w/]*\s*(set|export)\s+', flags=re.IGNORECASE)
ITEM_EXPRESSION = re.compile(r'[\w/]+')


def convert_ini_config_to_dict(config: ConfigParser) -> dict:
    """Translates a configparser object into a dict."""
    return {key: dict(value) for key, value in config.items()}


# DOTENV

def _sanitize_key_and_value(items: List[str]) -> List[str]:
    """Removes unwanted characters on key and value"""
    # if there are more than two items, we return the list unchanged, so an error will be raised later
    if len(items) != 2:
        return items
    key, value = items
    key = key.rstrip()
    value = value.lstrip()
    if (value.startswith('"') and value.endswith('"')) or (value.startswith("'") and value.endswith("'")):
        value = value[1:-1]
    return [key, value]


def get_dict_from_dotenv_file(filename: Union[Path, str]) -> Dict[str, str]:
    """
    :param filename: .env file where values are extracted.
    :return: a dict with keys and values extracted from the .env file.
    """
    result_dict = {}

    error_message = 'file {filename}: the line n°{index} is not correct: "{line}"'
    with open(filename) as f:
        for index, line in enumerate(f):
            stripped_line = line.strip()
            # we don't take into account comments
            if stripped_line.startswith('#'):
                continue
            # we don't take into account empty lines
            if not stripped_line:
                continue
            parts = stripped_line.split('#')  # we remove inline comments if there are any
            # we remove set or export command if there are any
            new_line = SET_EXPORT_EXPRESSION.sub('', parts[0].strip())

            # we get key and value
            parts = new_line.split('=')
            parts = _sanitize_key_and_value(parts)
            if len(parts) != 2 or ITEM_EXPRESSION.match(parts[0]) is None \
                    or ITEM_EXPRESSION.match(parts[1]) is None:
                line_number = index + 1
                raise DecodeError(message=error_message.format(filename=filename, index=line_number, line=new_line))
            result_dict[parts[0]] = parts[1]

    return result_dict


def bool_converter(value: str) -> bool:
    """
    :param value: a string to convert to bool.
    :return: False if lower case value in "0", "n", "no" and "false", otherwise, returns the value returned
    by the bool builtin function.
    """
    if value.lower() in ['n', '0', 'no', 'false']:
        return False
    return bool(value)


def string_list(value: str) -> List[str]:
    """
    :param value: a string to convert to a list of strings. Possible separators are space, ";", "," and ":".
    Note that separators other than space can be followed by one or more... spaces!
    :return: a list of strings.
    """
    return re.split(r',\s*|;\s*|:\s*|\s+', value)


def int_list(value: str) -> List[int]:
    """
    :param value: a string to convert to a list of int. Possible separators are space, ";", "," and ":".
    Note that separators other than space can be followed by one or more... spaces!
    :return: a list of int.
    """
    return [int(item) for item in string_list(value)]


def float_list(value: str) -> List[float]:
    """
    :param value: a string to convert to a list of float. Possible separators are space, ";", "," and ":".
    Note that separators other than space can be followed by one or more... spaces!
    :return: a list of float.
    """
    return [float(item) for item in string_list(value)]


def decimal_list(value: str) -> List[Decimal]:
    """
    :param value: a string to convert to a list of decimal.Decimal . Possible separators are space, ";", "," and ":".
    Note that separators other than space can be followed by one or more... spaces!
    :return: a list of decimal.Decimal .
    """
    return [Decimal(item) for item in string_list(value)]


def path_list(value: str) -> List[Path]:
    """
    :param value: a string to convert to a list of pathlib.Path . Possible separators are space, ";", "," and ":".
    Note that separators other than space can be followed by one or more... spaces!
    :return: a list of pathlib.Path .
    """
    return [Path(item) for item in string_list(value)]



class ConfigurorError(Exception):
    """base error"""


class DecodeError(ConfigurorError):
    def __init__(self, filename: str = 'the file', file_type: str = '', message: str = None):
        error_message = message or f'{filename} is not well {file_type} formatted'
        super().__init__(error_message)


class UnknownExtensionError(ConfigurorError):
    def __init__(self, extension: str = '', message: str = None):
        error_message = message or f'extension "{extension}" is not supported'
        super().__init__(error_message)



__version__ = '0.1.3'

from .main import Config, JSON_TYPE, YAML_TYPE, ENV_TYPE, TOML_TYPE, PYTHON_TYPE, INI_TYPE, EXTENSIONS
from .exceptions import ConfigurorError, DecodeError, UnknownExtensionError
from .utils import bool_converter, string_list, int_list, float_list, decimal_list, path_list

__all__ = [
    # main
    'Config', 'JSON_TYPE', 'YAML_TYPE', 'ENV_TYPE', 'TOML_TYPE', 'PYTHON_TYPE', 'INI_TYPE', 'EXTENSIONS',

    # exceptions
    'ConfigurorError', 'DecodeError', 'UnknownExtensionError',

    # utils
    'bool_converter', 'string_list', 'int_list', 'float_list', 'decimal_list', 'path_list'
]



"""Module which holds the Config class"""
import json
import os
from importlib import import_module
import importlib.util as import_util
from itertools import chain
from pathlib import Path
from typing import Dict, List, Callable, TypeVar, Union, Any

import yaml
from yaml.parser import ParserError as YamlParserError
import toml
# noinspection PyProtectedMember
from configparser import ConfigParser, Error as IniDecodeError, BasicInterpolation, ExtendedInterpolation

from .exceptions import DecodeError, UnknownExtensionError
from .utils import convert_ini_config_to_dict, get_dict_from_dotenv_file

Object = TypeVar('Object')

JSON_TYPE = 'json'

YAML_TYPE = 'yaml'

INI_TYPE = 'ini'

TOML_TYPE = 'toml'

PYTHON_TYPE = 'python'

ENV_TYPE = 'env'

EXTENSIONS = {
    JSON_TYPE: ['json'],
    YAML_TYPE: ['yml', 'yaml'],
    INI_TYPE: ['ini', 'cfg'],
    TOML_TYPE: ['toml'],
    PYTHON_TYPE: ['py'],
    ENV_TYPE: ['env']
}
# we create a list with all available extensions supported by configuror, it comes in handy
# for the implementation of load_from_files method
AVAILABLE_EXTENSIONS = list(chain(*[value for key, value in EXTENSIONS.items()]))


class Config(dict):

    def __init__(self, mapping_files: Dict[str, List[str]] = None, files: List[str] = None,
                 ignore_file_absence: bool = False, **kwargs):
        super(Config, self).__init__(**kwargs)
        self._type_error_message = '{filename} is not a string representing a path'
        self.load_from_mapping_files(mapping_files, ignore_file_absence)
        self.load_from_files(files, ignore_file_absence)

    @staticmethod
    def _path_is_ok(filename: str, ignore_file_absence: bool = False) -> bool:
        """
        :param filename: the filename to check for.
        :param ignore_file_absence: boolean flag to know if we raised an error if the file doesn't exist.
        :return: True if the file exists on the filesystem, False otherwise.
        """
        if os.path.isfile(filename):
            return True
        if ignore_file_absence:
            return False
        raise FileNotFoundError(f'file {filename} not found on the filesystem')

    def _filter_paths(self, filenames: List[str], ignore_file_absence: bool = False) -> List[str]:
        """
        :param filenames: list of files to check for.
        :param ignore_file_absence: boolean flag to know if we raised an error if a file in the list doesn't exist.
        :return: list of files guaranteed to exist on the filesystem.
        """
        for filename in filenames:
            if not isinstance(filename, str):
                raise TypeError(self._type_error_message.format(filename=filename))
        return [filename for filename in filenames if self._path_is_ok(filename, ignore_file_absence)]

    @staticmethod
    def getenv(key: str, default: Any = None, converter: Callable = None) -> Any:
        value = os.getenv(key, default)
        if converter is None:
            return value

        if not callable(converter):
            raise TypeError('cast must be a callable')
        return converter(value)

    def load_from_object(self, obj: Union[Object, str]) -> None:
        if isinstance(obj, str):
            obj = import_module(obj)
        for key in dir(obj):
            if key.isupper():
                self[key] = getattr(obj, key)

    def load_from_python_file(self, filename: str, ignore_file_absence: bool = False) -> bool:
        if not isinstance(filename, str):
            raise TypeError(self._type_error_message.format(filename=filename))

        if not self._path_is_ok(filename, ignore_file_absence):
            return False
        else:
            try:
                spec = import_util.spec_from_file_location(Path(filename).stem, filename)
                module = import_util.module_from_spec(spec)
                spec.loader.exec_module(module)
                self.load_from_object(module)
                return True
            except AttributeError:
                raise DecodeError(filename, PYTHON_TYPE)

    def load_from_json(self, filename: str, ignore_file_absence: bool = False) -> bool:
        if not self._path_is_ok(filename, ignore_file_absence):
            return False

        try:
            with open(filename) as f:
                self.update(json.load(f))
        except json.JSONDecodeError:
            raise DecodeError(filename, JSON_TYPE)
        return True

    def load_from_yaml(self, filename: str, ignore_file_absence: bool = False) -> bool:
        if not self._path_is_ok(filename, ignore_file_absence):
            return False

        try:
            with open(filename) as f:
                data = yaml.full_load(f)
                if not isinstance(data, dict):
                    return False
                self.update(data)
        except YamlParserError:
            raise DecodeError(filename, YAML_TYPE)
        return True

    def load_from_toml(self, filenames: Union[str, List[str]], ignore_file_absence: bool = False) -> bool:
        if not isinstance(filenames, (str, list)):
            raise TypeError('filenames must represent a path or list of paths')
        if isinstance(filenames, str):
            if not self._path_is_ok(filenames, ignore_file_absence):
                return False
            filenames = [filenames]

        filtered_filenames = self._filter_paths(filenames, ignore_file_absence)
        try:
            self.update(toml.load(filtered_filenames))
            return True
        except toml.TomlDecodeError:
            raise DecodeError(message=f'one of your files is not well {TOML_TYPE} formatted')
        except FileNotFoundError:  # This occurs when the list is empty, I just changed the error message to return
            raise FileNotFoundError(f'the list does not contain one {TOML_TYPE} valid file')

    def load_from_ini(self, filenames: Union[str, List], ignore_file_absence: bool = False,
                      interpolation_method: str = 'basic') -> bool:
        # we check interpolation method
        interpolation_error_message = 'interpolation_method must be either "basic" or "extended"'
        if not isinstance(interpolation_method, str):
            raise TypeError(interpolation_error_message)
        if interpolation_method.lower() not in ['basic', 'extended']:
            raise ValueError(interpolation_error_message)

        # we check filenames
        if not isinstance(filenames, (str, list)):
            raise TypeError('filenames must represent a path or list of paths')
        if isinstance(filenames, str):
            if not self._path_is_ok(filenames, ignore_file_absence):
                return False
            filenames = [filenames]

        filtered_filenames = self._filter_paths(filenames, ignore_file_absence)
        try:
            interpolation = ExtendedInterpolation() if interpolation_method.lower() == 'extended' \
                else BasicInterpolation()
            config = ConfigParser(interpolation=interpolation)
            config.read(filtered_filenames)
            self.update(convert_ini_config_to_dict(config))
            return True
        except IniDecodeError:
            raise DecodeError(message=f'one of your files is not well {INI_TYPE} formatted')

    def load_from_dotenv(self, filename: str, ignore_file_absence: bool = False) -> bool:
        if not self._path_is_ok(filename, ignore_file_absence):
            return False

        data = get_dict_from_dotenv_file(filename)
        if not data:
            return False

        for key, value in data.items():
            new_value = os.path.expandvars(value)
            os.environ[key] = new_value
            self[key] = new_value
        return True

    def load_from_mapping_files(self, mapping_files: Dict[str, List[str]] = None,
                                ignore_file_absence: bool = False) -> bool:
        if mapping_files is None:
            return False

        file_added = False
        for key, files in mapping_files.items():
            lower_key = key.lower()
            if lower_key not in EXTENSIONS.keys():
                raise UnknownExtensionError(key)
            if not isinstance(files, list):
                raise TypeError(f'{files} is not a list of files')

            existing_files = self._filter_paths(files, ignore_file_absence)
            if existing_files:  # if at least one file is added, the operation is considered realized
                file_added = True

                if lower_key == JSON_TYPE:
                    for file in existing_files:
                        self.load_from_json(file)

                if lower_key == YAML_TYPE:
                    for file in existing_files:
                        self.load_from_yaml(file)

                if lower_key == TOML_TYPE:
                    self.load_from_toml(existing_files)

                if lower_key == INI_TYPE:
                    self.load_from_ini(existing_files)

                if lower_key == PYTHON_TYPE:
                    for file in existing_files:
                        self.load_from_python_file(file)

                if lower_key == ENV_TYPE:
                    for file in existing_files:
                        self.load_from_dotenv(file)

        return file_added

    def load_from_files(self, filenames: List[str] = None, ignore_file_absence: bool = False) -> bool:
        if filenames is None:
            return False
        if not isinstance(filenames, list):
            raise TypeError(f'{filenames} is not a list of files')

        files = self._filter_paths(filenames, ignore_file_absence)
        if not files:
            return False
        else:
            for file in files:
                extension = file.split('.')[-1]
                if extension not in AVAILABLE_EXTENSIONS:
                    raise UnknownExtensionError(
                        message=f'{file} does not have a correct extension,'
                        f' supported extensions are: {AVAILABLE_EXTENSIONS}'
                    )

                if extension in EXTENSIONS[PYTHON_TYPE]:
                    self.load_from_python_file(file)

                elif extension in EXTENSIONS[JSON_TYPE]:
                    self.load_from_json(file)

                elif extension in EXTENSIONS[TOML_TYPE]:
                    self.load_from_toml(file)

                elif extension in EXTENSIONS[INI_TYPE]:
                    self.load_from_ini(file)

                elif extension in EXTENSIONS[YAML_TYPE]:
                    self.load_from_yaml(file)
                else:
                    self.load_from_dotenv(file)
            return True

    def get_dict_from_namespace(self, namespace: str, lowercase: bool = True,
                                trim_namespace: bool = True) -> Dict[str, Any]:
        result_dict = {}
        for key, value in self.items():
            if not isinstance(key, str) or not key.startswith(namespace):
                continue
            if trim_namespace:
                key = key[len(namespace):]
            if lowercase:
                key = key.lower()
            result_dict[key] = value

        return result_dict



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['configuror']

package_data = \
{'': ['*']}

install_requires = \
['pyyaml>=5.2,<6.0', 'toml>=0.10.0,<0.11.0']

setup_kwargs = {
    'name': 'configuror',
    'version': '0.2.0',
    'description': 'A configuration management toolkit',
    'long_description': "# configuror\n\n[![Pypi version](https://img.shields.io/pypi/v/configuror.svg)](https://pypi.org/project/configuror/)\n[![Build Status](https://travis-ci.com/lewoudar/configuror.svg?branch=master)](https://travis-ci.com/lewoudar/configuror)\n[![Windows Build Status](https://img.shields.io/appveyor/ci/lewoudar/configuror/master.svg?label=Windows)](https://ci.appveyor.com/project/lewoudar/configuror)\n[![Coverage Status](https://codecov.io/gh/lewoudar/configuror/branch/master/graphs/badge.svg?branch=master)](https://codecov.io/gh/lewoudar/configuror)\n[![Documentation Status](https://readthedocs.org/projects/configuror/badge/?version=latest)](https://configuror.readthedocs.io/en/latest/?badge=latest)\n[![License Apache 2](https://img.shields.io/hexpm/l/plug.svg)](http://www.apache.org/licenses/LICENSE-2.0)\n\nYour configuration management toolkit!\n\n## Why?\n\nWhile using [Flask](http://flask.pocoo.org/docs/1.0/), I realized that their Config class could be useful for any type\nof project. And the utility became more and more obvious to me when I looked at a project like\n[Ansible](https://docs.ansible.com/ansible/latest/index.html). If you look the \n[section](https://docs.ansible.com/ansible/latest/user_guide/playbooks_variables.html#variable-precedence-where-should-i-put-a-variable)\nwhere they define variable precedence, you will notice that there may be several locations for the configuration files,\nand these configuration files can be written in different formats (json, yaml..).\n\nWhat if there was a simple tool that would aggregate information from different sources in the order you wanted? This\nis **why** the configuror project exists!\n\n## Installation\n\n```bash\npip install configuror\n```\n\nconfiguror works starting from **python 3.6**. It has a few dependencies:\n- [pyyaml](https://pypi.org/project/PyYAML/) >= 5.1\n- [toml](https://pypi.org/project/toml/)\n\n## Documentation\n\nThe documentation is available at https://configuror.readthedocs.io/en/latest/.\n\n## Usage\n\nThe main class provided by configuror is `Config`. It is an extension of a regular dict object. There are two main ways \nto initialize it.\n\n### Using mapping files\n\n```python\nfrom configuror import Config\n\nmapping_files = {\n    'ini': ['foo.ini', 'bar.ini'],\n    'toml': ['foo.toml', 'bar.toml'],\n    'python': ['/path/to/python/file']\n}\nconfig = Config(mapping_files=mapping_files, ignore_file_absence=True)\n```\n\nYou can define a mapping of `file_type: <files>` where the `file_type` is the type of configuration file and `<files>` \nis the list of files from the lowest to the highest priority where values will be loaded.\n\nSince dictionaries are sorted starting from python3.6, the order of the keys is important as it will become the order of\nimportance of your files. For example in the example above, configuror will load values from files in the following order:\n- foo.ini\n- bar.ini\n- foo.toml\n- bar.toml\n- /path/to/python/file\n\nFor python files, only **uppercase** variables will be loaded.\n\nYou will notice the keyword argument `ignore_file_absence` in `Config` class initialization. If it is set to `True`, all\nfiles that does not exist will not raised `FileNotFoundError`. It comes in handy when you want to retrieve variables \nfrom files *that may or may not potentially exist*. By default this parameter is set to `False`.\n\nFile extension is not necessary when you use mapping files since the key is already telling which files we work with.\nThis is not the case with the second way to initialize `Config` class.\n\n### Using a list of files\n\n```python\nfrom configuror import Config\n\nfiles = [\n    'foo.yml',\n    'bar.toml',\n    'foobar.json',\n    '/path/to/python/file'\n    '.env'\n]\nconfig = Config(files=files)\n```\n\nIn this second form of initialization, you pass a list of files you want to retrieve values from the lowest to the\nhighest priority. File extension is **mandatory** here to help configuror to load the files properly.\n\nTo know file extensions supported by configuror, you can use the variable `EXTENSIONS`. it is a mapping\n`file_type: <extensions>` where `file_type` is a type of file supported like *yaml* and `extensions` is a list of\nrecognized extensions for this type of file, e.g: `[yml, yaml]`\n\nToday the file types supported are *toml*, *yaml*, *dotenv*, *ini*, *python* and *json*.\n\n### Other usages\n\nSince `Config` object is a dict-like object, you can pass arbitrary keyword arguments to initialize default values.\n\n```python\nfrom configuror import Config\n\nconfig = Config(FOO=2, BAR='a')\nprint(config)  # will print {'FOO': 2, 'BAR': 'a'}\n```\n\nYou can combine keyword arguments, mapping files and list of files at initialization. The order in which values will be\ninitialized is the following:\n- values from keyword arguments\n- values from mapping files\n- values from list of files\n\nYou can also add values from files after initialization. There are several practical methods for this:\n- `load_from_mapping_files(self, mapping_files: Dict[str, List[str]], ignore_file_absence: bool)`: It is in fact the \nmethod used under the hood when you initialized `Config` object by passing the parameter `mapping_files`.\n\n- `load_from_files(self, files: List[str], ignore_file_absence: bool)`: It is the method used under the hood when you\ninitialized `Config` objects by passing the parameter `files`. \n\n- `load_from_object(self, obj: Union[Object, str])`: `obj` can be an object or a path to a project module\n(with dotted notation). Only uppercase attributes of the corresponding object will be retrieved.\n\n- `load_from_python_file(self, filename: str, ignore_file_absence: bool)`: Loads values from an arbitrary python\nfile. It would be preferable if it were not a file related to your project (i.e a module). Only uppercase variables\nare considered.\n\n- `load_from_json(self, filename: str, ignore_file_absence: bool)`: Loads values from a json file.\n\n- `load_from_yaml(self, filename: str, ignore_file_absence: bool)`: Loads values from a yaml file.\n\n- `load_from_toml(self, filenames: Union[str, List[str]], ignore_file_absence: bool)`: Loads values from a toml file\nor a list of toml files.\n\n- `load_from_ini(self, filenames: Union[str, List], ignore_file_absence: bool, interpolation_method: str = 'basic')`:\nLoads values from an ini file or a list of ini files. There are two interpolation methods that can be used: **basic** \nor **extended** like explained in the\n[documentation](https://docs.python.org/3/library/configparser.html#interpolation-of-values).\n\n- `load_from_dotenv(self, filename: str, ignore_file_absence: bool)`: Loads values from a dotenv file.\n\nBonus: You also have the `update` method of a dict to add/update values.\n",
    'author': 'lewoudar',
    'author_email': 'lewoudar@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://configuror.readthedocs.io/en/stable',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
